/*
c:/devkitGP2X/bin/arm-linux-gcc colourd.c -o colourd.gpe_ -static -lm -g && cp colourd.gpe_ colourd.gpe colourd f:/
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>
#include <float.h>
#include <math.h>

#define RATIO		0.33333333f
#define DICHO_ERR_RANGE	0.1f

int sig3=0;

typedef struct
{
	float x;
	float y;

} point_t;

typedef struct
{
	point_t p0;	// names match to the points described there http://en.wikipedia.org/wiki/B%C3%A9zier_curve
	point_t p1;
	point_t p2;
	point_t p3;

	point_t q0;
	point_t q1;
	point_t q2;
	point_t r0;
	point_t r1;
	point_t b;

} bezier_t;

void handle_sigquit(int sig)
{
	sig3 = 1;
}

void nsleep(uint32_t s, uint32_t ns)	//simplifies calls to nanosleep
{
	struct timespec tspec;

	tspec.tv_sec = (time_t) s;
	tspec.tv_nsec = ns;

	nanosleep(&tspec, NULL);
}

float dist(point_t a, point_t b)
{
	return sqrtf((a.x*a.x)+(b.x*b.x));
}

float find_point_y(bezier_t c, float x)
{
	// we try to find the point in curve "c" at x coordinate "x" to return its y coordinate
	// we do so by a kind of weighted dichotomy
	float t;
	float exp_x;	// expected x

	t = (x - c.p0.x) / (c.p3.x - c.p0.x);		// weighted c.position where c.b.x is expected to be found to be close to

	while (fabsf(c.b.x - x) > DICHO_ERR_RANGE)
	{
		exp_x = c.p0.x * (1.0f - t) + c.p3.x * t;	// this is where we expect c.b.x to be found

		c.q0.x = c.p0.x * (1.0f - t) + c.p1.x * t;
		c.q1.x = c.p1.x * (1.0f - t) + c.p2.x * t;
		c.q2.x = c.p2.x * (1.0f - t) + c.p3.x * t;
		c.r0.x = c.q0.x * (1.0f - t) + c.q1.x * t;
		c.r1.x = c.q1.x * (1.0f - t) + c.q2.x * t;
		c.b.x  = c.r0.x * (1.0f - t) + c.r1.x * t;

		t = (t * (x - c.p0.x)) / (c.b.x - c.p0.x);	// make t be something else so that the next c.b.x can be closer to x
	}

	c.q0.y = c.p0.y * (1.0f - t) + c.p1.y * t;
	c.q1.y = c.p1.y * (1.0f - t) + c.p2.y * t;
	c.q2.y = c.p2.y * (1.0f - t) + c.p3.y * t;
	c.r0.y = c.q0.y * (1.0f - t) + c.q1.y * t;
	c.r1.y = c.q1.y * (1.0f - t) + c.q2.y * t;
	c.b.y  = c.r0.y * (1.0f - t) + c.r1.y * t;

	return c.b.y;
}

int main()
{
	FILE *config, *profile;
	int32_t devgpio = open("/dev/GPIO", O_RDONLY);
	int32_t devmem = open("/dev/mem", O_RDWR);
	uint16_t *screen16, *memregs16;
	uint32_t *memregs32;
	uint32_t i, ip, ix;
	uint8_t lut[256]={0,};
	point_t *point;
	int32_t point_count;
	bezier_t *curve;
	point_t o0, o1, o2, h1, h2;
	float da, db, dt;	// distances
	float foo;
	float end_level;	// x coordinate of the end point

	memregs32 = (uint32_t *) mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, devmem, 0xc0000000);
	memregs16 = (uint16_t *) memregs32;

	/********FILE I/O********/
	config = fopen("colourd", "r");

	if (!config)	// if the profile file is absent
	{
		fprintf(stderr, "Error : configuration file missing. Now quitting.\n");
		exit(EXIT_FAILURE);
	}

	fscanf(config, "%d", &point_count);	// read how many points there are

	point = malloc(point_count * sizeof(point_t));	// allocate the points array
	curve = calloc((point_count+1), sizeof(bezier_t));	// allocate the bezier curves array
printf("DEBUG 0 : %d\n", (uint32_t) curve[0].p0.x+1);

	for (ip=0; ip<point_count; ip++)
	{
		point[ip].x = (float) (ip+1) * (256.0f / (float) (point_count+1));	// initialize x coordinates
		fscanf(config, "\n%f", &point[ip].y);	// read each y coordinate into the array
	}

	fscanf(config, "\n%f", &end_level);	// read the end point
	fclose (config);
	/********END FILE I/O********/

	curve[point_count].p3.x = 255.0f;	// initialize curve end points
	curve[point_count].p3.y = end_level;
	curve[point_count].p2 = curve[point_count].p3;
printf("DEBUG 1 : %d\n", (uint32_t) curve[0].p0.x+1);

	for (ip=0; ip<point_count; ip++)	// calculate control points
	{
		o1 = point[ip];
		curve[ip].p3 = o1;
		curve[ip+1].p0 = o1;
printf("DEBUG 2 : %d\n", (uint32_t) curve[0].p0.x+1);

		if (ip==0)
		{
			o0.x=0.0f;
			o0.y=0.0f;
		}
		else
			o0 = point[ip-1];
		if (ip==point_count-1)
		{
			o2.x=255.0f;
			o2.y=end_level;
		}
		else
			o2 = point[ip+1];

		da = dist(o0, o1);
		db = dist(o1, o2);

		h1.x = ((o2.x-o0.x) * (-da / (da+db)) * RATIO) + o1.x;
		h1.y = ((o2.y-o0.y) * (-da / (da+db)) * RATIO) + o1.y;
		h2.x = ((o2.x-o0.x) * (db / (da+db)) * RATIO) + o1.x;
		h2.y = ((o2.y-o0.y) * (db / (da+db)) * RATIO) + o1.y;

		curve[ip].p2 = h1;
		curve[ip+1].p1 = h2;
	}

	for (ip=0; ip<point_count+1; ip++)	// calculate curve data
	{
printf("DEBUG 3 : %d\n", (uint32_t) curve[0].p0.x+1);
		for (ix = (uint32_t) curve[ip].p0.x+1; ix < (uint32_t) curve[ip].p3.x+1; ix++)
		{
			//printf("%3d\t%3.0f\n", ix, find_point_y(curve[ip], (float) ix));
			if (ix!=256)
				lut[ix] = (uint8_t) (find_point_y(curve[ip], (float) ix) + 0.5f);
		}
	}

	while (sig3==0)	// to stop this loop, send a SIGQUIT signal to the process (kill -3)
	{
		memregs16[0x2880 >> 1] &= ~(1<<12);	//makes the bit #12 be 0 in order to activate gamma correction
		for (i=0; i<256; i++)
		{
			memregs16[0x295C >> 1] = i << 1;	// indicate which input value we're dealing with
	
			memregs16[0x295E >> 1] = (lut[i] << 8) | lut[i];	// filling the register with the found value
			memregs16[0x295E >> 1] = lut[i];
		}

		signal(SIGQUIT, handle_sigquit);
		nsleep(5, 0);	// do it every 5 seconds
	}

	return 0;
}
